# Copyright 1999-2006 University of Chicago
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

use Globus::GRAM::Error;
use Globus::GRAM::JobState;
use Globus::GRAM::JobManager;
use Globus::Core::Paths;

# NOTE: This package name must match the name of the .pm file!!
package Globus::GRAM::JobManager::pbs;

@ISA = qw(Globus::GRAM::JobManager);

my ($mpirun, $mpiexec, $qsub, $qstat, $qdel, $cluster, $cpu_per_node, $remote_shell);

BEGIN
{
    $mpiexec        = (defined($ENV{MPIEXEC}) ? $ENV{MPIEXEC} : "mpiexec");
    $mpirun         = (defined($ENV{MPIRUN}) ? $ENV{MPIRUN} : "mpirun");
    $qsub           = (defined($ENV{QSUB}) ? $ENV{QSUB} : "qsub");
    $qstat          = (defined($ENV{QSTAT}) ? $ENV{QSTAT} : "qstat");
    $qdel           = (defined($ENV{QDEL}) ? $ENV{QDEL} : "qdel");
    $cluster        = (defined($ENV{CLUSTER}) ? $ENV{CLUSTER} : 1);
    $cpu_per_node   = (defined($ENV{CPU_PER_NODE}) ? $ENV{CPU_PER_NODE} : 1);
    $remote_shell   = (defined($ENV{REMOTE_SHELL}) ? $ENV{REMOTE_SHELL} : "ssh");
    $softenv_dir    = (defined($ENV{SOFTENV_DIR}) ? $ENV{SOFTENV_DIR} : "");
    $soft_msc       = "$softenv_dir/bin/soft-msc";
    $softenv_load   = "$softenv_dir/etc/softenv-load.sh";

}

sub myceil ($)
{
    my $x = shift;
    ( abs($x-int($x)) < 1E-12 ) ? int($x) : int($x < 0 ? $x : $x+1.0);
}

sub submit
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $status;
    my $pbs_job_script;
    my $pbs_job_script_name;
    my $pbs_qsub_err_name ;
    my $errfile = '';
    my $job_id;
    my $rsh_env;
    my @arguments;
    my $email_when = '';
    my $args;
    my $cache_pgm = "$Globus::Core::Paths::bindir/globus-gass-cache";

    $self->log("Entering pbs submit");

    # Reject jobs that want streaming, if so configured
    if ( $description->streamingrequested() &&
	 $description->streamingdisabled() ) {

	$self->log("Streaming is not allowed.");
	return Globus::GRAM::Error::OPENING_STDOUT;
    }

    # check jobtype
    if(defined($description->jobtype()))
    {
        if($description->jobtype !~ /^(mpi|single|multiple)$/)
        {
            return Globus::GRAM::Error::JOBTYPE_NOT_SUPPORTED;
        }
    }
    if( $description->directory eq '')
    {
        return Globus::GRAM::Error::RSL_DIRECTORY();
    }
    if ($description->directory() =~ m|^[^/]|) {
        $description->add("directory",
                $ENV{HOME} . '/' . $description->directory());
    }
    chdir $description->directory() or
        return Globus::GRAM::Error::BAD_DIRECTORY();

    $self->nfssync( $description->executable() )
        unless $description->executable() eq '';
    $self->nfssync( $description->stdin() )
        unless $description->stdin() eq '';
    if( $description->executable eq '')
    {
        return Globus::GRAM::Error::RSL_EXECUTABLE();
    }
    elsif(! -f $description->executable())
    {
        return Globus::GRAM::Error::EXECUTABLE_NOT_FOUND();
    }
    elsif(! -x $description->executable())
    {
        return Globus::GRAM::Error::EXECUTABLE_PERMISSIONS();
    }
    elsif( $description->stdin() eq '')
    {
        return Globus::GRAM::Error::RSL_STDIN;
    }
    elsif(! -r $description->stdin())
    {
       return Globus::GRAM::Error::STDIN_NOT_FOUND();
    }

    $self->log("Determining job max time cpu from job description");
    if(defined($description->max_cpu_time())) 
    {
        $cpu_time = $description->max_cpu_time();
        $self->log("   using maxcputime of $cpu_time");
    }
    elsif(! $cluster && defined($description->max_time()))
    {
        $cpu_time = $description->max_time();
        $self->log("   using maxtime of $cpu_time");
    }
    else
    {
        $cpu_time = 0;
        $self->log('   using queue default');
    }

    $self->log("Determining job max wall time limit from job description");
    if(defined($description->max_wall_time()))
    {
        $wall_time = $description->max_wall_time();
        $self->log("   using maxwalltime of $wall_time");
    }
    elsif($cluster && defined($description->max_time()))
    {
       $wall_time = $description->max_time();
       $self->log("   using maxtime of $wall_time");
    }
    else
    {
        $wall_time = 0;
        $self->log('   using queue default');
    }

    $self->log('Building job script');

    $pbs_job_script_name = $self->job_dir() . '/scheduler_pbs_job_script';

    local(*JOB);
    open( JOB, '>' . $pbs_job_script_name );
    print JOB<<"EOF";
#! /bin/sh
# PBS batch job script built by Globus job manager
#
#PBS -S /bin/sh
EOF

    if($description->name() ne '')
    {
        print JOB '#PBS -N ', $description->name(), "\n";
    }
    if($description->email_address() ne '')
    {
        print JOB '#PBS -M ', $description->email_address(), "\n";
    }
    if($description->emailonabort() eq 'yes')
    {
        $email_when .= 'a';
    }
    if($description->emailonexecution() eq 'yes')
    {
        $email_when .= 'b';
    }
    if($description->emailontermination() eq 'yes')
    {
        $email_when .= 'e';
    }
    if($email_when eq '')
    {
        $email_when = 'n';
    }
    print JOB "#PBS -m $email_when\n";

    if($description->queue() ne '')
    {
        print JOB '#PBS -q ', $description->queue(), "\n";
    }
    if($description->project() ne '')
    {
        print JOB '#PBS -A ', $description->project(), "\n";
    }

    if($cpu_time != 0)
    {
        if($description->jobtype() eq 'multiple')
        {
            if ($description->totalprocesses() > 0)
            {
                $total_cpu_time = $cpu_time * $description->totalprocesses();
            }
            else
            {
                $total_cpu_time = $cpu_time * $description->count();
            }
        }
        else
        {
            $total_cpu_time = $cpu_time;
        }
        print JOB "#PBS -l pcput=${cpu_time}:00\n";
        print JOB "#PBS -l cput=${total_cpu_time}:00\n";
    }

    if($wall_time != 0)
    {
        print JOB "#PBS -l walltime=${wall_time}:00\n";
    }

    if($description->max_memory() != 0)
    {
        if($description->jobtype() eq 'multiple')
        {
            if ($description->totalprocesses() > 0)
            {
                $max_memory = $description->max_memory()
                            * $description->totalprocesses();
            }
            else
            {
                $max_memory = $description->max_memory()
                            * $description->count();
            }
        }
        else
        {
            $max_memory = $description->max_memory();
        }
        print JOB "#PBS -l mem=${max_memory}mb\n";
    }
    print JOB '#PBS -o ', $description->stdout(), "\n";
    print JOB '#PBS -e ', $description->stderr(), "\n";

    if (defined $description->nodes())
    {
        #Generated by ExtensionsHandler.pm from resourceAllocationGroup elements
        print JOB '#PBS -l nodes=', $description->nodes(), "\n";
    }
    elsif($description->host_count() != 0)
    {
        print JOB '#PBS -l nodes=', $description->host_count(), "\n";
    }
    elsif($cluster && $cpu_per_node != 0)
    {
        print JOB '#PBS -l nodes=',
        myceil($description->count() / $cpu_per_node), "\n";
    }

    ### SoftEnv extension ###
    if ($softenv_dir ne '')
    {
        $self->setup_softenv(
            $self->job_dir() . '/pbs_softenv_job_script',
            $soft_msc,
            $softenv_load,
            *JOB);
    }
    #########################

    $rsh_env = '';

    foreach my $tuple ($description->environment())
    {
        if(!ref($tuple) || scalar(@$tuple) != 2)
        {
            return Globus::GRAM::Error::RSL_ENVIRONMENT();
        }

        push(@new_env, $tuple->[0] . '="' . $tuple->[1] . '"');

        $tuple->[0] =~ s/\\/\\\\/g;
        $tuple->[0] =~ s/\$/\\\$/g;
        $tuple->[0] =~ s/"/\\\"/g; #"
        $tuple->[0] =~ s/`/\\\`/g; #`

        $tuple->[1] =~ s/\\/\\\\/g;
        $tuple->[1] =~ s/\$/\\\$/g;
        $tuple->[1] =~ s/"/\\\"/g; #"
        $tuple->[1] =~ s/`/\\\`/g; #`

        $rsh_env .= $tuple->[0] . '="' . $tuple->[1] . "\";\n"
                 .  'export ' . $tuple->[0] . ";\n";
    }

    print JOB "$rsh_env";

    print JOB "\n#Change to directory requested by user\n";
    print JOB 'cd ' . $description->directory() . "\n";

    @arguments = $description->arguments();

    foreach(@arguments)
    {
        if(ref($_))
        {
            return Globus::GRAM::Error::RSL_ARGUMENTS;
        }
    }
    if($#arguments >= 0)
    {
        foreach(@arguments)
        {
            $self->log("Transforming argument \"$_\"\n");
            $_ =~ s/\\/\\\\/g;
            $_ =~ s/\$/\\\$/g;
            $_ =~ s/"/\\\"/g; #"
            $_ =~ s/`/\\\`/g; #`
            $self->log("Transformed to \"$_\"\n");

            $args .= '"' . $_ . '" ';
        }
    }
    else
    {
        $args = '';
    }

    if ($description->executable() =~ m|^[^/]|)
    {
        $description->add('executable', './' . $description->executable());
    }
    if($description->jobtype() eq 'multiple' && (($description->count()==1) || !$cluster))
    {
        my $process_count;
        if ($description->totalprocesses() > 0)
        {
            $process_count = $description->totalprocesses();
        }
        else
        {
            $process_count = $description->count();
        }

        print JOB "pids=''\n";
        print JOB "exit_code=0\n";
        for(my $i = 0; $i < $process_count; $i++)
        {
            print JOB $description->executable(), " $args <",
                $description->stdin(), "&\n";
            print JOB "pids=\"\$pids \$!\"\n";
        }
        print JOB <<EOF;
            for x in \$pids; do
                wait \$x
                tmp_exit_code=\$?
                if [ \$exit_code = 0 -a \$tmp_exit_code != 0 ]; then
                    exit_code=\$tmp_exit_code
                fi
            done
            exit \$exit_code
EOF
    }
    elsif($description->jobtype() eq 'mpi' ||
            $description->jobtype() eq 'multiple')
    {
        my $count;
        if ($description->totalprocesses() > 0)
        {
            $count = $description->totalprocesses();
        }
        else
        {
            $count = $description->count();
        }
        my $cmd_script_name ;
        my $cmd_script ;
        my $stdin = $description->stdin();

        $cmd_script_name = $self->job_dir() . '/scheduler_pbs_cmd_script';

        local(*CMD);
        if ( open( CMD, ">$cmd_script_name" ) ) 
        {
            print CMD "#!/bin/sh\n";

            ### SoftEnv extension ###
            $self->setup_softenv(
                $self->job_dir() . '/pbs_softenv_cmd_script',
                $soft_msc,
                *CMD);
            #########################

            print CMD 'cd ', $description->directory(), "\n";
            print CMD "$rsh_env\n";
            print CMD $description->executable(), " $args\n";
            close(CMD);
            chmod 0700, $cmd_script_name;

            $self->nfssync( $cmd_script_name );
        } 
        else 
        {
            # FIXME: what to do in error?
            return Globus::GRAM::Error::TEMP_SCRIPT_FILE_FAILED();
        }

        if ($description->jobtype() eq "mpi")
        {
            if ($mpiexec ne 'no')
            {
                my $machinefilearg = "";
                if ($cluster)
                {
                    $machinefilearg = ' -machinefile $PBS_NODEFILE';
                }
                if ($description->totalprocesses() > 0)
                {
                    print JOB "$mpiexec $machinefilearg -n "
                            . $description->totalprocesses();
                }
                else
                {
                    print JOB "$mpiexec $machinefilearg -n "
                            . $description->count();
                }
            }
            else
            {
                if ($description->totalprocesses() > 0)
                {
                    print JOB "$mpirun -np " . $description->totalprocesses();
                }
                else
                {
                    print JOB "$mpirun -np " . $description->count();
                }
                if ($cluster)
                {
                    print JOB ' -machinefile $PBS_NODEFILE';
                }
            }

            print JOB " $cmd_script_name < " .  $description->stdin() . "\n";
        }
        else
        {
            my $exit_prefix=$self->job_dir() . '/exit';

            print JOB <<"EOF";

hosts=\`cat \$PBS_NODEFILE\`;
counter=0
while test \$counter -lt $count; do
    for host in \$hosts; do
        if test \$counter -lt $count; then
            $remote_shell \$host "/bin/sh $cmd_script_name; echo \\\$? > $exit_prefix.\$counter" < $stdin &
            counter=\`expr \$counter + 1\`
        else
            break
        fi
    done
done
wait

counter=0
exit_code=0
while test \$counter -lt $count; do
    /bin/touch $exit_prefix.\$counter;

    read tmp_exit_code < $exit_prefix.\$counter
    if [ \$exit_code = 0 -a \$tmp_exit_code != 0 ]; then
        exit_code=\$tmp_exit_code
    fi
    counter=\`expr \$counter + 1\`
done

exit \$exit_code
EOF
        }
    }
    else
    {
        print JOB $description->executable(), " $args <",
            $description->stdin(), "\n";
    }
    close(JOB);

    $pbs_qsub_err_name = $self->job_dir() . '/scheduler_pbs_submit_stderr';
    $errfile = "2>$pbs_qsub_err_name";

    $self->nfssync( $pbs_job_script_name );
    $self->nfssync( $pbs_qsub_err_name );
    $self->nfssync( $description->stdout, 1 );
    $self->nfssync( $description->stderr, 1 );
    $self->log("submitting job -- $qsub < $pbs_job_script_name $errfile");
    chomp($job_id = `$qsub < $pbs_job_script_name $errfile`);

    if($? == 0)
    {
        $self->log("job submission successful, setting state to PENDING");
        return {JOB_ID => $job_id,
                JOB_STATE => Globus::GRAM::JobState::PENDING };
    }
    else
    {
        local(*ERR);
        open(ERR, "<$pbs_qsub_err_name");
        local $/;
        my $stderr = <ERR>;
        close(ERR);

        $self->log("qsub returned $job_id");
        $self->log("qsub stderr $stderr");

        open(ERR, ">" . $description->stderr());
        print ERR $stderr;
        close(ERR);

        $stderr =~ s/\n/\\n/g;

        $self->respond({GT3_FAILURE_MESSAGE => $stderr });
    }

    return Globus::GRAM::Error::JOB_EXECUTION_FAILED();
}

sub poll
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $job_id = $description->jobid();
    my $state;
    my $status_line;
    my $exit_code;

    $self->log("polling job $job_id");

    # Get job id from the full qstat output.
    $_ = (grep(/job_state/, $self->pipe_out_cmd($qstat, '-f', $job_id)))[0];
    # get the exit code of the qstat command.  for info search $CHILD_ERROR
    # in perlvar documentation.
    $exit_code = $? >> 8;

    $self->log("qstat job_state line is: $_");

    # return code 153 = "Unknown Job Id".
    # verifying that the job is no longer there.
    if($exit_code == 153)
    {
        $self->log("qstat rc is 153 == Unknown Job ID == DONE");
        $state = Globus::GRAM::JobState::DONE;
        $self->nfssync( $description->stdout() )
            if $description->stdout() ne '';
        $self->nfssync( $description->stderr() )
            if $description->stderr() ne '';
    }
    else
    {

        # Get 3rd field (after = )
        $_ = (split(/\s+/))[3];

        if(/Q|W|T/)
        {
            $state = Globus::GRAM::JobState::PENDING;
        }
        elsif(/S|H/)
        {
            $state = Globus::GRAM::JobState::SUSPENDED
        }
        elsif(/R|E/)
        {
            $state = Globus::GRAM::JobState::ACTIVE;
        }
        elsif(/C/)
        {
            $state = Globus::GRAM::JobState::DONE;
            $self->nfssync( $description->stdout() )
                if $description->stdout() ne '';
            $self->nfssync( $description->stderr() )
                if $description->stderr() ne '';
        }
        else
        {
            # This else is reached by an unknown response from pbs.
            # It could be that PBS was temporarily unavailable, but that it
            # can recover and the submitted job is fine.
            # So, we want the JM to ignore this poll and keep the same state
            # as the previous state.  Returning an empty hash below will tell
            # the JM to ignore the respose.
            $self->log("qstat returned an unknown response.  Telling JM to ignore this poll");
            return {};
        }
    }

    return {JOB_STATE => $state};
}

sub cancel
{
    my $self = shift;
    my $description = $self->{JobDescription};
    my $job_id = $description->jobid();

    $self->log("cancel job $job_id");

    $self->fork_and_exec_cmd( $qdel, $job_id );

    if($? == 0)
    {
        return { JOB_STATE => Globus::GRAM::JobState::FAILED }
    }

    return Globus::GRAM::Error::JOB_CANCEL_FAILED();
}

1;
